import utilities as ut
import numpy as np
from scipy.spatial import ConvexHull
from scipy.spatial import Delaunay
import fileprocessing as fp
import copy
import plot2d


# avhengig av at platform har z=0
def get_first_layer(nd, coords):
    arr = []
    for point in coords:
        if point[nd - 1] == 1:
            arr.append(point)
    return arr


# hull er av typen ConvexHull, returneres pga plotting
# front er en liste av alle hjørnepnktene til polyhedronen hull
def calculate_convex_front(region):
    points = np.asarray(region)
    hull = ConvexHull(points)
    front = []
    for index in hull.vertices:
        front.append(region[index])
    return front


def find_shadowed(obj_coords, front, processed_voxels):
    voxels_inside_hull = []
    shadowed = []
    for v in obj_coords:
        if in_hull(v, front):
            voxels_inside_hull.append(v)
    for v in voxels_inside_hull:
        if v not in processed_voxels and v not in shadowed:
            shadowed.append(v)
    return shadowed


def asn(nd, v1, v2):
    x = ut.v_subtr(v1, v2)
    if ut.n_norm(2, nd, x) == 1 or ut.n_norm(2, nd, x) == 2 ** (1 / 2):
        return 1
    else:
        return 0


def generate_list_of_ASNs(nd, voxel, obj_coords):
    neighbours = []
    for potential_neighbour in obj_coords:
        if asn(nd, voxel, potential_neighbour):
            neighbours.append(potential_neighbour)
    return neighbours


# fra https://stackoverflow.com/questions/16750618/whats-an-efficient-way-to-find-if-a-point-lies-in-the-convex-hull-of-a-point-cl
# returns a boolean array where True indicates points that lie in the given convex hull
# modification: from return hull.find_simplex(p) >= 0
def in_hull(p, front):
    if not isinstance(front, Delaunay):
        front = Delaunay(front)
    return front.find_simplex(p) >= 0


def find_next_layer_greedy(obj_coords, current_layer, processed_voxels, current_front, nd):
    next_layer = []
    for n, point in enumerate(current_layer):
        neighbours = generate_list_of_ASNs(nd, point, obj_coords)
        for neighbour in neighbours:

            if not in_hull(neighbour, current_front):
                if neighbour not in processed_voxels and neighbour not in next_layer:
                    next_layer.append(neighbour)
    return next_layer


def convex_front_advancing(nd, method, filename, filepath):
    obj, platform = fp.load_file(nd, filename, filepath)

    current_layer = get_first_layer(nd, obj)
    shadowed_voxels = []
    previous_front = []
    growing_field = []
    processed_voxels = []

    i = 0
    print("Generating layers")
    while current_layer:
        i += 1
        print("Generating layer", i, "...")

        ut.add_to(current_layer, processed_voxels)
        next_layer, current_front, shadowed_voxels = greedy(nd, platform, obj, current_layer,
                                                            processed_voxels, previous_front)
        if method == "isp":
            next_layer = incremental_shadow_prevention(current_front, obj, platform, processed_voxels,
                                                       next_layer, shadowed_voxels)

        previous_front = current_front
        growing_field.append(current_layer)
        current_layer = next_layer

    return growing_field, shadowed_voxels


# platf_coords : list of platform coordinates
# obj_coords : list of object coordinates
# nd : number of dimensions [2, 3]
# method : string, which method to use ("greedy", "greedy_sp")
def greedy(nd, platf_coords, obj_coords, current_layer, processed_voxels, previous_front):
    processed_region = ut.union(3, previous_front, current_layer, platf_coords, [])
    current_front = calculate_convex_front(processed_region)
    shadowed_voxels = find_shadowed(obj_coords, current_front, processed_voxels)
    next_layer = find_next_layer_greedy(obj_coords, current_layer, processed_voxels, current_front, nd)

    return next_layer, current_front, shadowed_voxels


def incremental_shadow_prevention(old_front, model, platform, processed, next_layer, prev_shadowed):
    potential_shadowed = []
    reduced_next_layer = []
    of = copy.deepcopy(old_front)

    un = ut.union(3, old_front, platform, next_layer, [])
    new_front = calculate_convex_front(un)

    new_processed = []
    for p in processed:
        new_processed.append(p)

    for p in next_layer:
        new_processed.append(p)

    potential_shadowed = find_shadowed(model, new_front, new_processed)
    # Hvis det nye layeret genererer shadowed totalt, men ikke flere enn tidligere (dvs ingen nye voxels er shadowed)
    if len(potential_shadowed) <= len(prev_shadowed):
        return next_layer
    else:
        print("Trying to improve layer ...")
        q = copy.deepcopy(next_layer)
        test_processed = []
        j = 0
        while q:
            j += 1
            p = q[0]
            q.pop(0)
            u = ut.union(4, p, platform, reduced_next_layer, of)

            test_front = calculate_convex_front(u)

            for point in processed:
                if point not in test_processed:
                    test_processed.append(point)

            test_processed.append(p)
            test_shadowed = find_shadowed(model, test_front, test_processed)

            if len(test_shadowed) < len(potential_shadowed):
                reduced_next_layer.append(p)

        if reduced_next_layer:
            next_layer = reduced_next_layer

        return next_layer
